"""Wrapper to convert a ChatArena environment into a PettingZoo compatible environment."""
from __future__ import annotations

import functools
from typing import Any, Dict, Optional

import pettingzoo
from gymnasium import spaces
from pettingzoo.utils.env import AgentID, ObsType

import chatarena
from chatarena.arena import Arena

import string

CHAR_SET = string.printable

class PettingZooCompatibilityV0(pettingzoo.AECEnv):
    """This compatibility wrapper converts a ChatArena environment into a PettingZoo environment.

    ChatArena (or Chat Arena) is a Multi-Agent Language Game Environments for LLMs. The goal is to develop communication and collaboration capabilities of AIs.
    """

    metadata = {
        "render_modes": ["human"],
        "name": "PettingZooCompatibilityV0",
        "is_parallelizable": False,
    }

    def __init__(
        self,
        env: chatarena.arena.Arena | None = None,
        arena_name: str | None = None,
        string_observation: bool | None = True,
        max_turns: int | None = 25,
        render_mode: str | None = None,
    ):
        """Wrapper to convert a OpenSpiel environment into a PettingZoo environment.

        Args:
            env (chatarena.arena.Arena): chatarena arena to wrap
            arena_name (Optional[str]): chatarena environment to load from file (e.g., "examples/chameleon.json")
            max_turns (Optional[int]): maximum number of turns before environment truncates
            render_mode (Optional[str]): rendering mode
        """
        super().__init__()

        if env is not None:
            self._env = env
        elif arena_name is not None:
            self._env = Arena.from_config(arena_name)
        else:
            raise ValueError("Arena not specified, please us env or arena_name arguments.")

        self._env.reset() # this resets the underlying arena as well as each player

        self.possible_agents = list(self._env.name_to_player.keys())
        self.name_to_player_mapping = self._env.name_to_player


        self.string_observation = string_observation
        self.max_turns = max_turns
        self.render_mode = render_mode

        self.terminations = {}
        self.truncations = {}
        self.rewards = {}
        self.infos = {a: {} for a in self.possible_agents}


    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent: AgentID):
        """observation_space.

        We get the observation space from the underlying environment.
        Args:
            agent (AgentID): agent
        """
        # TODO: finalize obs space (dicts with agent name may not be necessary)
        observation_space = spaces.Dict(
            {
                agent: spaces.Text(max_length=256, min_length=0, charset=CHAR_SET)
                for agent in self.possible_agents
            }
        )
        return observation_space

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent: AgentID):
        """action_space.

        Get the action space from the underlying environment.

        Args:
            agent (AgentID): agent

        Returns:
            space
        """
        # TODO: finalize action space (this enables agents to send messages to specific other players)
        action_space = spaces.Dict(
            {
                agent: spaces.Text(max_length=256, min_length=0, charset=CHAR_SET)
                for agent in self.possible_agents
            }
        )
        return action_space

    def render(self):
        """render.

        Print the current game state.
        """
        if not hasattr(self, "initial_timestep"):
            raise UserWarning(
                "You must reset the environment using reset() before calling render()."
            )

        self._env.environment.print()
        pass

    def observe(self, agent: AgentID) -> ObsType:
        """observe.

        Args:
            agent (AgentID): agent (e.g., "Player 1")

        Returns:
            observation
        """
        messages = self._env.environment.get_observation(agent) # this will only return the messages this agent can see
        if len(messages) > 0:
            self.current_turn = messages[-1].turn
        else:
            self.current_turn = 0
        new_messages = [m for m in messages if m.turn == self.current_turn] # we only send the current timestep messages

        # string observation
        if self.string_observation == True:
            observation = ""
            for m in new_messages:
                observation += f"{m.agent_name}: {m.content}"

        # dict observation
        else:
            observation = {m.agent_name: m.content for m in new_messages}

        self.infos[agent]["obs_dict"] = {m.agent_name: m.content for m in new_messages}

        return observation


    def close(self):
        """close."""
        pass

    def _unravel_timestep(self, timestep: chatarena.arena.TimeStep):
        # get observation
        messages = timestep.observation
        if len(messages) > 0:
            self.current_turn = messages[-1].turn
        else:
            self.current_turn = 0
        new_messages = [m for m in messages if m.turn == self.current_turn]  # we only send the current timestep messages

        # string observation
        if self.string_observation == True:
            observation = ""
            for m in new_messages:
                observation += f"{m.agent_name}: {m.content}"

        # dict observation
        else:
            observation = {m.agent_name: m.content for m in new_messages}


        # get rewards
        rewards = timestep.reward

        # get termination
        termination = timestep.terminal

        # get truncation
        truncation = self.current_turn > self.max_turns

        # get info
        player_idx = self.possible_agents.index(self.agent_selection)
        player_obj = self._env.players[player_idx]
        info = {"turn": self.current_turn, "global_prompt": player_obj.global_prompt, "agent_desc": player_obj.role_desc}

        return observation, rewards, termination, truncation, info

    def reset(
        self,
        return_info: bool | None = False,
        seed: int | None = None,
        options: dict | None = None,
    ):
        """reset.

        Args:
            seed (Optional[int]): seed
            options (Optional[Dict]): options
        """

        if seed is not None:
            print("WARNING: seeding is not supported for LLMs.")

        # reset the chat arena environment
        self.initial_timestep = self._env.reset()
        self.turn = 0

        # get the first player
        self.agent_selection = self._env.environment.get_next_player()

        observation, reward, termination, truncation, info = self._unravel_timestep(self.initial_timestep)

        agent = self.agent_selection
        self.rewards = reward
        self.terminations[agent] = termination
        self.truncations[agent] = truncation
        self.infos[agent] = info

        # all agents
        self.agents = self.possible_agents[:]

        # boilerplate stuff
        self._cumulative_rewards = {a: 0 for a in self.agents}
        self.rewards = self.initial_timestep.reward
        self.terminations = {a: False for a in self.agents}
        self.truncations = {a: False for a in self.agents}

    def step(self, action: str):
        """Steps.

        Steps the agent with an action.

        Args:
            action (str): action
        """
        if (
            self.terminations[self.agent_selection]
            or self.truncations[self.agent_selection]
        ):
            return self._was_dead_step(action)

        agent = self.agent_selection
        timestep = self._env.environment.step(player_name=agent, action=action)

        observation, reward, termination, truncation, info = self._unravel_timestep(timestep)

        self.rewards = reward
        self.terminations[agent] = termination
        self.truncations[agent] = truncation
        self.infos[agent] = info

        self._cumulative_rewards[agent] = 0

        self.agent_selection = self._env.environment.get_next_player()

        self._accumulate_rewards()

        if self.render_mode == "human":
            self.render()






from .base import IntelligenceBackend
from ..config import BackendConfig


# An Error class for the human backend
class HumanBackendError(Exception):
    def __init__(self, agent_name: str):
        self.agent_name = agent_name
        super().__init__(f"Human backend requires a UI to get input from {agent_name}.")


class Human(IntelligenceBackend):
    stateful = False
    type_name = "human"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def to_config(self) -> BackendConfig:
        return BackendConfig(backend_type=self.type_name)

    def query(self, agent_name: str, **kwargs) -> str:
        raise HumanBackendError(agent_name)



from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.styles import Style
from rich.console import Console
from rich.text import Text
from rich.color import ANSI_COLOR_NAMES
import random

from ..arena import Arena, TooManyInvalidActions
from ..backends.human import HumanBackendError

ASCII_ART = r"""
_________  .__               __      _____                                   
\_   ___ \ |  |__  _____   _/  |_   /  _  \  _______   ____    ____  _____   
/    \  \/ |  |  \ \__  \  \   __\ /  /_\  \ \_  __ \W/ __ \  /    \ \__  \  
\     \____|   Y  \ / __ \_ |  |  /    |    \ |  | \/\  ___/ |   |  \ / __ \_
 \______  /|___|  /(____  / |__|  \____|__  / |__|    \___  >|___|  /(____  /
        \/      \/      \/                \/              \/      \/      \/ 
"""

visible_colors = [color for color in ANSI_COLOR_NAMES.keys() if
                  color not in ["black", "white", "red", "green"] and "grey" not in color]

MAX_STEPS = 5

import logging

# Set logging level to ERROR
logging.getLogger().setLevel(logging.ERROR)


class ArenaCLI:
    """
    The CLI user interface for ChatArena.
    """

    def __init__(self, arena: Arena):
        self.arena = arena

    def launch(self, max_steps: int = None, interactive: bool = True):
        """
        Run the CLI
        """
        if not interactive and max_steps is None:
            max_steps = MAX_STEPS

        console = Console()
        # Print ascii art
        console.print(ASCII_ART, style="bold dark_orange3")
        timestep = self.arena.reset()
        console.print("🏟 Chat Arena Initialized!", style="bold green")

        env = self.arena.environment
        players = self.arena.players

        env_desc = self.arena.global_prompt
        num_players = env.num_players
        player_colors = random.sample(visible_colors, num_players)  # sample different colors for players
        name_to_color = dict(zip(env.player_names, player_colors))
        # System and Moderator messages are printed in red
        name_to_color["System"] = "red"
        name_to_color["Moderator"] = "red"

        console.print(f"[bold green underline]Environment ({env.type_name}) description:[/]\n{env_desc}")

        # Print the player name, role_desc and backend_type
        for i, player in enumerate(players):
            player_name = Text(f"[{player.name} ({player.backend.type_name})] Role Description:")
            player_name.stylize(f"bold {name_to_color[player.name]} underline")
            console.print(player_name)
            console.print(player.role_desc)

        console.print("\n========= Arena Start! ==========\n", style="bold green")

        step = 0
        while not timestep.terminal:
            if interactive:
                command = prompt([('class:command', "command (n/r/q/s/h) > ")],
                                 style=Style.from_dict({'command': 'blue'}),
                                 completer=WordCompleter(
                                     ['next', 'n', 'reset', 'r', 'exit', 'quit', 'q', 'help', 'h', 'save', 's']))
                command = command.strip()

                if command == "help" or command == "h":
                    console.print("Available commands:")
                    console.print("    [bold]next or n or <Enter>[/]: next step")
                    console.print("    [bold]exit or quit or q[/]: exit the game")
                    console.print("    [bold]help or h[/]: print this message")
                    console.print("    [bold]reset or r[/]: reset the game")
                    console.print("    [bold]save or s[/]: save the history to file")
                    continue
                elif command == "exit" or command == "quit" or command == "q":
                    break
                elif command == "reset" or command == "r":
                    timestep = self.arena.reset()
                    console.print("\n========= Arena Reset! ==========\n", style="bold green")
                    continue
                elif command == "next" or command == "n" or command == "":
                    pass
                elif command == "save" or command == "s":
                    # Prompt to get the file path
                    file_path = prompt([('class:command', "save file path > ")],
                                       style=Style.from_dict({'command': 'blue'}))
                    file_path = file_path.strip()
                    # Save the history to file
                    self.arena.save_history(file_path)
                    # Print the save success message
                    console.print(f"History saved to {file_path}", style="bold green")
                else:
                    console.print(f"Invalid command: {command}", style="bold red")
                    continue

            try:
                timestep = self.arena.step()
            except HumanBackendError as e:
                # Handle human input and recover with the game update
                human_player_name = env.get_next_player()
                if interactive:
                    human_input = prompt(
                        [('class:user_prompt', f"Type your input for {human_player_name}: ")],
                        style=Style.from_dict({'user_prompt': 'ansicyan underline'})
                    )
                    # If not, the conversation does not stop
                    timestep = env.step(human_player_name, human_input)
                else:
                    raise e  # cannot recover from this error in non-interactive mode
            except TooManyInvalidActions as e:
                # Print the error message
                console.print(f"Too many invalid actions: {e}", style="bold red")
                break

            # The messages that are not yet logged
            messages = [msg for msg in env.get_observation() if not msg.logged]
            # Print the new messages
            for msg in messages:
                message_text = Text(f"[{msg.agent_name}->{msg.visible_to}]: {msg.content}")
                message_text.stylize(f"bold {name_to_color[msg.agent_name]}", 0,
                                     len(f"[{msg.agent_name}->{msg.visible_to}]:"))
                console.print(message_text)
                msg.logged = True

            step += 1
            if max_steps is not None and step >= max_steps:
                break

        console.print("\n========= Arena Ended! ==========\n", style="bold red")



from ..config import BackendConfig

from .base import IntelligenceBackend
from .openai import OpenAIChat
from .cohere import CohereAIChat
from .human import Human
from .hf_transformers import TransformersConversational
from .anthropic import Claude

ALL_BACKENDS = [
    Human,
    OpenAIChat,
    CohereAIChat,
    TransformersConversational,
    Claude,
]

BACKEND_REGISTRY = {backend.type_name: backend for backend in ALL_BACKENDS}


# Load a backend from a config dictionary
def load_backend(config: BackendConfig):
    try:
        backend_cls = BACKEND_REGISTRY[config.backend_type]
    except KeyError:
        raise ValueError(f"Unknown backend type: {config.backend_type}")

    backend = backend_cls.from_config(config)
    return backend



from chatarena.environments.base import Environment, TimeStep
from chatarena.message import Message, MessagePool
from typing import List, Dict, Union
from chatarena.agent import Player
from chatarena.backends import OpenAIChat
from chatarena.arena import Arena
from chatarena.utils import extract_code, extract_jsons
from io import StringIO
import sys
import traceback

class PythonREPL:
    """Simulates a standalone Python REPL."""
    def __init__(self):
        self.globals = {}

    def run(self, command: str) -> str:
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        try:
            exec(command, self.globals)
            sys.stdout = old_stdout
            output = mystdout.getvalue()
        except Exception:
            sys.stdout = old_stdout
            output = traceback.format_exc()
        return output


class IterativeCoding(Environment):
    type_name = "coding"

    def __init__(self, task:str=""):
        super().__init__(player_names=["coder", "verifier"])

        self.task = task
        # The "state" of the environment is maintained by the message pool
        self.message_pool = MessagePool()
        self.phase = "code" # "code", "verify", "iterate"
        self.python_repl = PythonREPL()
        self.max_turns = 10
        self._terminal = False
        self.reset()
        self.last_code = ""

    def get_next_player(self) -> str:
        if self.phase == "code":
            return "coder"
        elif self.phase == "iterate":
            return "coder"
        elif self.phase == "verify":
            return "verifier"

    def _moderator_speak(self, text: str, visible_to: Union[str, List[str]] = "all"):
        """
        moderator say something
        """
        message = Message(agent_name="Moderator", content=text, turn=self.turn, visible_to=visible_to)
        self.message_pool.append_message(message)

    def reset(self):
        self.turn = 0
        self.message_pool.reset()
        self._moderator_speak(f"For the following task \n ```{self.task}```. "
                              f"\n Write some testcases and then an actual function that implement the task. Everything should be in a single code block", visible_to="coder")
        observation = self.get_observation(self.get_next_player())
        self._terminal = False
        self.turn += 1
        return TimeStep(observation=observation, reward=self.get_zero_rewards(), terminal=self._terminal)

    def get_observation(self, player_name=None) -> List[Message]:
        if player_name is None:
            return self.message_pool.get_all_messages()
        else:
            return self.message_pool.get_visible_messages(player_name, turn=self.turn + 1)

    def process_broken(self):
        self._moderator_speak(f"The process is broken. Please restart the game.")
        self._terminal = True
        observation = self.get_observation(self.get_next_player())
        return TimeStep(observation=observation, reward=self.get_zero_rewards(), terminal=self._terminal)

    def step(self, player_name: str, action: str) -> TimeStep:
        assert player_name == self.get_next_player(), f"Wrong player! It is {self.get_next_player()} turn."
        visible_to = "all"
        message = Message(agent_name=player_name, content=action, turn=self.turn, visible_to=visible_to)
        self.message_pool.append_message(message)
        if self.phase in ["iterate", "code"]:
            code_list = extract_code(action)
            if len(code_list) != 1:
                return self.process_broken()
            self.last_code = code_list[0]
            interpreter_output = self.python_repl.run(code_list[0])
            self.phase = "verify"
        elif self.phase == "verify":
            json_list = extract_jsons(action)
            if len(json_list) != 1:
                return self.process_broken()
            if json_list[0]["result"] == "correct":
                self._terminal = True
                self._moderator_speak(f"Tests passed! Here's the code: \n ```{self.last_code}```")
                return TimeStep(observation=self.get_observation(self.get_next_player()),
                                reward=self.get_one_rewards(),
                                terminal=True)
            self.phase = "iterate"


        if self.phase == "verify":
            self._moderator_speak(f"Here's the outputs: {interpreter_output}. Is the code correct? Output with json format.",
                                  visible_to="verifier")
        elif self.phase == "iterate":
            self._moderator_speak(f"Now iterate your code with feedbacks. First think about why and then write the new code.", visible_to="coder")

        self.turn += 1
        return TimeStep(observation=self.get_observation(self.get_next_player()),
                        reward=self.get_zero_rewards(),
                        terminal=self._terminal)


if __name__ == "__main__":
    coder_role_description = """
    You are a coder. You are going to follow a workflow of coding to implement a specific function.
    Your implementation will be tested by the verifier. If the implementation is wrong, you will try output new implementation given the feedback.
    Your output can include your reasoning process but the code part should always be surrounded by triple backticks.
    """

    verifier_role_description = """
    You are a verifier. You are going to verify if the code is correct or not according to the interpretor outputs.
    You should always output a json with following format:
    { 
    "outputs_extraction": the outputs from the interpreter output showing the error or correctness of the code,
    "result": "correct" or "incorrect",
    }
    """

    task = """
    Write a python function for detecting if there's a json within a bunch of text.
    The input of this function is a string, and the output is a boolean.
    If there are multiple jsons in the string, return True if any of them is valid.
    """

    coder = Player("coder", role_desc=coder_role_description,
                   backend=OpenAIChat(max_tokens=1024, model="gpt-4"))
    verifier = Player("verifier", role_desc=verifier_role_description,
                        backend=OpenAIChat(max_tokens=1024, model="gpt-4"))
    env = IterativeCoding(task=task)
    arena = Arena([coder, verifier], env)
    arena.launch_cli()



import re
from pettingzoo.classic import tictactoe_v3

from chatarena.environments.base import Environment, TimeStep
from typing import List, Union

from ..message import Message, MessagePool


def action_string_to_action(action: str) -> int:
    pattern = r"(X|O): \((\d), (\d)\)"
    match = re.match(pattern, action)

    if not match:
        return -1
    items = [item for item in match.groups()]
    coords = [int(coord) for coord in items[1:]]
    row, column = coords
    if row not in [1, 2, 3]:
        return -1
    if column not in [1, 2, 3]:
        return -1

    row = row - 1
    column = column - 1
    return row + column * 3


class PettingzooTicTacToe(Environment):
    type_name = "pettingzoo:tictactoe"

    def __init__(self, player_names: List[str], **kwargs):
        super().__init__(player_names=player_names, **kwargs)
        self.env = tictactoe_v3.env()

        # The "state" of the environment is maintained by the message pool
        self.message_pool = MessagePool()
        self._terminal = False
        self.reset()

    def reset(self):
        self.env.reset()
        self.current_player = 0
        self.turn = 0
        self.message_pool.reset()

        obs_dict, reward, terminal, truncation, info = self.env.last()
        observation = self.get_observation()
        self._terminal = terminal
        return TimeStep(observation=observation, reward=reward, terminal=terminal)

    def get_next_player(self) -> str:
        return self.player_names[self.current_player]

    def get_observation(self, player_name=None) -> List[Message]:
        if player_name is None:
            return self.message_pool.get_all_messages()
        else:
            return self.message_pool.get_visible_messages(player_name, turn=self.turn + 1)

    def _moderator_speak(self, text: str, visible_to: Union[str, List[str]] = "all"):
        """
        moderator say something
        """
        message = Message(agent_name="Moderator", content=text, turn=self.turn, visible_to=visible_to)
        self.message_pool.append_message(message)

    def is_terminal(self) -> bool:
        return self._terminal

    def step(self, player_name: str, action: str) -> TimeStep:
        assert player_name == self.get_next_player(), f"Wrong player! It is {self.get_next_player()} turn."

        message = Message(agent_name=player_name, content=action, turn=self.turn)
        self.message_pool.append_message(message)
        # Convert the action to the AlphaZero format
        action_index = action_string_to_action(action)
        if action_index == -1:
            raise ValueError(f"Invalid action: {action}")

        self.env.step(action_index)
        obs_dict, reward, terminal, truncation, info = self.env.last()

        self._terminal = terminal  # Update the terminal state
        reward = reward  # TODO: bug here, reward needs to be a dict

        self.current_player = 1 - self.current_player
        self.turn += 1
        self._moderator_speak("\n" + self.render_ansi(obs_dict["observation"]))

        return TimeStep(observation=self.get_observation(), reward=reward, terminal=terminal)

    def check_action(self, action: str, agent_name: str) -> bool:
        # This can be implemented depending on how you want to validate actions for a given agent
        action_index = action_string_to_action(action)
        if action_index == -1:
            return False
        elif self.env.last()[0]["action_mask"][action_index] == 0:
            return False
        else:
            return True

    def render_ansi(self, observation):
        string = ""
        observation = observation.transpose(1, 0, 2)
        for row in observation:
            string += "|"
            for column in row:
                symbol = "_"
                if column[self.current_player] == 1:
                    symbol = "X"
                elif column[1 - self.current_player] == 1:
                    symbol = "O"
                string += " " + symbol + " |"
            string += "\n"
        return string

    def print(self):
        obs_dict, reward, terminal, truncation, info = self.env.last()
        print(self.render_ansi(obs_dict["observation"]))



import numpy as np
from typing import List, Dict, Union
from chatarena.agent import Player
from chatarena.backends import OpenAIChat, Claude
from langchain.document_loaders import OnlinePDFLoader

from chatarena.environments.base import Environment, TimeStep
from chatarena.message import Message, MessagePool
from chatarena.agent import SIGNAL_END_OF_CONVERSATION
from chatarena.arena import Arena
from chatarena.utils import is_json_inside

DEFAULT_ORDER_BOOK = {
    "bids": [
        {"price": 3.98, "amount": 862},
        {"price": 3.99, "amount": 562},
        {"price": 4.00, "amount": 431},
    ],
    "asks": [
        {"price": 4.02, "amount": 12},
        {"price": 4.03, "amount": 285},
        {"price": 4.04, "amount": 210},
    ]
}


class Trading(Environment):
    type_name = "trading"

    def __init__(self, doc: str = ""):
        super().__init__(player_names=["researcher", "manager", "trader"])

        self.doc = doc
        # The "state" of the environment is maintained by the message pool
        self.message_pool = MessagePool()
        self.phase = "research"  # "research", "discussion", "trading"
        self._terminal = False
        self.reset()

    def reset(self):
        self.current_player = "researcher"
        self.turn = 0
        self.message_pool.reset()

        self._moderator_speak(f"Here's the whitepaper of a new cryptocurrency. Please read it carefully:\n {self.doc}",
                              visible_to="researcher")
        observation = self.get_observation(self.get_next_player())
        self._terminal = False
        self.phase = "discussion"
        return TimeStep(observation=observation,
                        reward=self.get_zero_rewards(),
                        terminal=self._terminal)

    def get_next_player(self) -> str:
        if self.phase == "research":
            return "researcher"
        elif self.phase == "discussion":
            if self.current_player == "researcher":
                return "manager"
            elif self.current_player == "manager":
                return "researcher"
        elif self.phase == "trading":
            return "trader"
        else:
            raise ValueError(f"Unknown phase {self.phase}")

    def get_observation(self, player_name=None) -> List[Message]:
        if player_name is None:
            return self.message_pool.get_all_messages()
        else:
            return self.message_pool.get_visible_messages(player_name, turn=self.turn + 1)

    def _moderator_speak(self, text: str, visible_to: Union[str, List[str]] = "all"):
        """
        moderator say something
        """
        message = Message(agent_name="Moderator", content=text, turn=self.turn, visible_to=visible_to)
        self.message_pool.append_message(message)

    def is_terminal(self) -> bool:
        return self._terminal

    def step(self, player_name: str, action: str) -> TimeStep:
        assert player_name == self.get_next_player(), f"Wrong player! It is {self.get_next_player()} turn."
        message = Message(agent_name=player_name, content=action, turn=self.turn)
        self.message_pool.append_message(message)
        if self.phase == "trading":
            self._terminal = True
        if is_json_inside(action) and self.phase == "discussion" and player_name == "manager":
            self.phase = "trading"
            self._moderator_speak(f"Here's the order book please put orders \n{DEFAULT_ORDER_BOOK}",
                                  visible_to="trader")

        self.turn += 1
        self.current_player = self.get_next_player()
        return TimeStep(observation=self.get_observation(self.get_next_player()),
                        reward=self.get_zero_rewards(),
                        terminal=self._terminal)


if __name__ == "__main__":
    researcher_role_description = """
    You are a researcher for crypto-trading.
    You are going to analyse the whitepaper of a new cryptocurrency.
    After finishing the reading, you'll dicuss with a trader, helping him to make a decision.
    """

    manager_role_description = """
    You are managing a crypto fund.
    You are going to discuss with a researcher about a new cryptocurrency called impt.io whose market cap is 5.2 million dollars.
    Try to figure out all the information you need to make a decision.
    Try to ask at least 3 round of questions before you make the decision.
    When you are ready to make the decision, output a json with the following format:
    { 
    "reasong": the reason for your decision,
    "decision": "long" or "short"",
    }
    """

    trader_role_description = """
    You are a trader for crypto-fund.
    You are going to follow the "long" or "short" decision of your manager and put limit orders accorrding to the information in the order book.
    If the decision is "long", you should put a buy order (amount be position) otherwise you should put a sell order (amount be negative).
    The price should always be in favor of the fund this means if the decision is "long", the price should be lower than the lowest ask price in the order book,
    otherwise the price should be higher than highest ask price in the order book.
    You should a output a json with the following format:
    {
    "orders": [
        {"price": price of the order, "amount": amount to buy or sell. positive means buy, negative means sell},
    ]
    }   
    """

    loader = OnlinePDFLoader("https://impt.io/assets/documents/whitepaper/en.pdf")
    doc = loader.load()

    researcher = Player(name="researcher", role_desc=researcher_role_description,
                        global_prompt="", backend=Claude(max_tokens=1024, model="claude-v1.3-100k"))
    manager = Player(name="manager", role_desc=manager_role_description,
                     global_prompt="", backend=OpenAIChat(max_tokens=1024, model="gpt-4"))
    trader = Player(name="trader", role_desc=trader_role_description,
                    global_prompt="", backend=OpenAIChat(max_tokens=1024))
    env = Trading(doc=str(doc))
    arena = Arena([researcher, manager, trader], env)
    arena.launch_cli()



import re
import json

def is_json(myjson):
    try:
        json_object = json.loads(myjson)
    except ValueError as e:
        return False
    return True

def is_json_inside(text):
    text = re.sub('\s+', ' ', text)
    matches = re.findall(r'\{.*?\}', text)
    for match in matches:
        if is_json(match):
            return True
    return False

def extract_jsons(text):
    text = re.sub('\s+', ' ', text)
    matches = re.findall(r'\{.*?\}', text)
    parsed_jsons = []
    for match in matches:
        try:
            json_object = json.loads(match)
            parsed_jsons.append(json_object)
        except ValueError as e:
            pass
    return parsed_jsons


def extract_code(text):
    text = re.sub('```python', '```', text)
    matches = re.findall(r'```(.*?)```', text, re.DOTALL)
    parsed_codes = []
    for match in matches:
        parsed_codes.append(match)
    return parsed_codes


class AttributedDict(dict):
    """
    A dict class whose keys are automatically set as attributes of the class.
    Serializable to JSON.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError

    def __delattr__(self, key):
        del self[key]

    # check whether the key is string when adding the key
    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise ValueError("The key must be a string")
        super().__setitem__(key, value)

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            self[key] = value



from typing import List
from abc import abstractmethod

from ..config import BackendConfig, Configurable
from ..message import Message


class IntelligenceBackend(Configurable):
    """An abstraction of the intelligence source of the agents."""
    stateful = None
    type_name = None

    @abstractmethod
    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # registers the arguments with Configurable

    def __init_subclass__(cls, **kwargs):
        # check if the subclass has the required attributes
        for required in ('stateful', 'type_name',):
            if getattr(cls, required) is None:
                raise TypeError(f"Can't instantiate abstract class {cls.__name__} without {required} attribute defined")
        return super().__init_subclass__(**kwargs)

    def to_config(self) -> BackendConfig:
        self._config_dict["backend_type"] = self.type_name
        return BackendConfig(**self._config_dict)

    @abstractmethod
    def query(self, agent_name: str, role_desc: str, history_messages: List[Message], global_prompt: str = None,
              request_msg: Message = None, *args, **kwargs) -> str:
        raise NotImplementedError

    @abstractmethod
    async def async_query(self, agent_name: str, role_desc: str, history_messages: List[Message],
                          global_prompt: str = None, request_msg: Message = None, *args, **kwargs) -> str:
        """Async querying"""
        raise NotImplementedError

    # reset the state of the backend
    def reset(self):
        if self.stateful:
            raise NotImplementedError
        else:
            pass



from typing import List, Dict, Union
import uuid
import json
import csv
import logging

from .agent import Player
from .environments import Environment, TimeStep, load_environment
from .backends import Human
from .config import ArenaConfig


class TooManyInvalidActions(Exception):
    pass


class Arena:
    """
    Utility class that manages the game environment and players
    """

    def __init__(self, players: List[Player], environment: Environment, global_prompt: str = None):
        # Create a container for the players and environment and reset the game
        self.players = players
        self.environment = environment
        self.global_prompt = global_prompt

        self.current_timestep = environment.reset()
        self.uuid = uuid.uuid4()  # Generate a unique id for the game
        self.invalid_actions_retry = 5

    @property
    def num_players(self):
        return self.environment.num_players

    @property
    def name_to_player(self) -> Dict[str, Player]:
        return {player.name: player for player in self.players}

    def reset(self) -> TimeStep:
        # Reset the environment
        self.current_timestep = self.environment.reset()
        # Reset the players
        for player in self.players:
            player.reset()
        # Reset the uuid
        self.uuid = uuid.uuid4()
        return self.current_timestep

    def step(self) -> TimeStep:
        """
        Take a step in the game: one player takes an action and the environment updates
        """
        player_name = self.environment.get_next_player()
        player = self.name_to_player[player_name]  # get the player object
        observation = self.environment.get_observation(player_name)  # get the observation for the player

        timestep = None
        for i in range(self.invalid_actions_retry):  # try to take an action for a few times
            action = player(observation)  # take an action
            if self.environment.check_action(action, player_name):  # action is valid
                timestep = self.environment.step(player_name, action)  # update the environment
                break
            else:  # action is invalid
                logging.warning(f"{player_name} made an invalid action {action}")
                continue

        if timestep is None:  # if the player made invalid actions for too many times, terminate the game
            warning_msg = f"{player_name} has made invalid actions for {self.invalid_actions_retry} times. Terminating the game."
            logging.warning(warning_msg)
            raise TooManyInvalidActions(warning_msg)

        return timestep

    def next_is_human(self):
        """
        check if the next player is human
        """
        player_name = self.environment.get_next_player()
        player = self.name_to_player[player_name]
        return isinstance(player.backend, Human)

    def run(self, num_steps: int = 1):
        """
        run the game for num_turns
        """
        for i in range(num_steps):
            timestep = self.step()
            if timestep.terminal:
                break

    @classmethod
    def from_config(cls, config: Union[str, ArenaConfig]):
        """
        create an arena from a config
        """
        # If config is a path, load the config
        if isinstance(config, str):
            config = ArenaConfig.load(config)

        global_prompt = config.get("global_prompt", None)

        # Create the players
        players = []
        for player_config in config.players:
            # Add public_prompt to the player config
            if global_prompt is not None:
                player_config["global_prompt"] = global_prompt

            player = Player.from_config(player_config)
            players.append(player)

        # Check that the player names are unique
        player_names = [player.name for player in players]
        assert len(player_names) == len(set(player_names)), "Player names must be unique"

        # Create the environment
        config.environment["player_names"] = player_names  # add the player names to the environment config
        env = load_environment(config.environment)

        return cls(players, env, global_prompt=global_prompt)

    def to_config(self) -> ArenaConfig:
        """
        convert the arena to a config
        """
        # return {
        #     "players": [player.to_config() for player in self.players],
        #     "environment": self.environment.to_config(),
        #     "global_prompt": self.global_prompt
        # }
        return ArenaConfig(
            players=[player.to_config() for player in self.players],
            environment=self.environment.to_config(),
            global_prompt=self.global_prompt
        )

    def launch_cli(self, max_steps: int = None, interactive: bool = True):
        """
        launch the command line interface
        """
        from chatarena.ui.cli import ArenaCLI
        cli = ArenaCLI(self)
        cli.launch(max_steps=max_steps, interactive=interactive)

    def save_config(self, path: str):
        """
        save the config to a file
        """
        config = self.to_config()
        config.save(path)

    def save_history(self, path: str):
        """
        save the history of the game to a file
        Supports csv and json formats.
        """
        messages = self.environment.get_observation()
        message_rows = []

        if path.endswith(".csv"):
            header = ["agent_name", "content", "turn", "timestamp", "visible_to", "msg_type"]
            for message in messages:
                message_row = [
                    message.agent_name,
                    message.content,
                    message.turn,
                    str(message.timestamp),
                    message.visible_to,
                    message.msg_type,
                ]
                message_rows.append(message_row)

            with open(path, "w") as f:
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows(message_rows)
        elif path.endswith(".json"):
            for message in messages:
                message_row = {
                    "agent_name": message.agent_name,
                    "content": message.content,
                    "turn": message.turn,
                    "timestamp": str(message.timestamp),
                    "visible_to": message.visible_to,
                    "msg_type": message.msg_type,
                }
                message_rows.append(message_row)

            with open(path, "w") as f:
                json.dump(message_rows, f, indent=4)
        else:
            raise ValueError("Invalid file format")



from setuptools import setup, find_packages

_deps = [
    "openai==0.27.2",
    "anthropic==0.2.8",
    "cohere==4.3.1",
    "transformers>=4.27.4",
    "tenacity==8.2.2",
    "gradio==3.20.0",
    "rich==13.3.3",
    "prompt_toolkit==3.0.38",
    "pettingzoo==1.23.0",
    "chess==1.9.4",
]

with open("README.md", "r") as f:
    long_description = f.read()

requirements = _deps

setup(
    name="chatarena",
    version="0.1.10",
    author="Yuxiang Wu",
    author_email="yuxiang.cs@gmail.com",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chatarena/chatarena",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.7",
    install_requires=requirements,
)



from .base import Environment, TimeStep
from .conversation import Conversation, ModeratedConversation
from .chameleon import Chameleon
from .pettingzoo_chess import PettingzooChess
from .pettingzoo_tictactoe import PettingzooTicTacToe

from ..config import EnvironmentConfig

ALL_ENVIRONMENTS = [
    Conversation,
    ModeratedConversation,
    Chameleon,
    PettingzooChess,
    PettingzooTicTacToe,
]

ENV_REGISTRY = {env.type_name: env for env in ALL_ENVIRONMENTS}


# Load an environment from a config dictionary
def load_environment(config: EnvironmentConfig):
    try:
        env_cls = ENV_REGISTRY[config["env_type"]]
    except KeyError:
        raise ValueError(f"Unknown environment type: {config['env_type']}")

    env = env_cls.from_config(config)
    return env



from chatarena.agent import Player, Moderator
from chatarena.backends import OpenAIChat
from chatarena.backends.human import Human
from chatarena.arena import Arena
from chatarena.environments.conversation import ModeratedConversation, Conversation

MODEL = "gpt-4"


def main():
    # Describe the environment (which is shared by all players)
    environment_description = """
    This is a board of advisors that advices the CEO of a startup on a question that the CEO ask.
    The board of advisors is composed of six members with different expertise.
    1. Industry veteran in finance like Warren Buffet
    2. Industry veteran in business strategy like Jeff Bezos
    3. Industry veteran in marketing like Seth Godin
    4. Industry veteran in negotiation like Chris Voss
    5. Industry veteran in technology like Elon Musk
    The five board members have to discuss among them. They are free to disagree with each other, and suggest an alternative approach, until they reach consensus.
    Do not always agree with the CEO or the other advisors on the board.
    """

    ceo = Player(name="CEO", backend=Human(),
                 role_desc="You are CEO.",
                 # terminal_condition="Have the board of advisors reach consensus? Answer yes or no.",
                 global_prompt=environment_description)

    warrent_buffet = """Warren Buffet follows the Benjamin Graham school of value investing, which looks for securities whose prices are unjustifiably low based on their intrinsic worth. He has developed several core tenets to help him employ his investment philosophy to maximum effect. These tenets fall into four categories: business, management, financial measures, and value.

In terms of business tenets, Buffet restricts his investments to businesses he can easily analyze. In terms of management tenets, Buffet evaluates the track records of a company’s higher-ups to determine if they have historically reinvested profits back into the company or if they have redistributed funds to back shareholders in the form of dividends. In terms of financial measures, Buffet focuses on low-levered companies with high profit margins. Finally, in terms of value tenets, Buffet looks for companies with a special product and good profit margins."""
    player1 = Player(name="Finance Advisor", backend=OpenAIChat(model=MODEL),
                     role_desc=f"You are the finance advisor like Warrent Buffet. Here is a brief description of Warrent Buffet:\n {warrent_buffet}",
                     global_prompt=environment_description)

    jeff_bezos = """Jeff Bezos is known for his success as an investor and businessman. He manages his portfolio through the investment firm he founded, Bezos Expeditions, and currently holds positions in dozens of companies. Some of the important tips to invest like Jeff Bezos include building a diversified portfolio, being a long-term investor, and investing in modern, cutting-edge companies ². He also believes in finding opportunity in crisis and knowing what the crowd thinks. """
    player2 = Player(name="Business Strategist", backend=OpenAIChat(model=MODEL),
                     role_desc=f"You are the business strategist like Jeff Bezos. Here is a brief description of Jeff Bezos:\n {jeff_bezos}",
                     global_prompt=environment_description)

    seth_godin = """Seth Godin is a bestselling author and entrepreneur known for his insights on marketing. He advises entrepreneurs to build products worth shouting about, rather than shouting about their products from the rooftops. He recommends approaching marketing strategy with four key points of focus: Coordination, Trust, Permission, and the Exchange of Ideas. He also emphasizes the importance of spreading your idea, thinking out of the box, and making your customers obsessed with your product or service."""
    player3 = Player(name="Marketing Expert", backend=OpenAIChat(model=MODEL),
                     role_desc=f"You are the marketing expert like Seth Godin. Here is a brief description of Seth Godin:\n{seth_godin}",
                     global_prompt=environment_description)

    christ_voss = """Chris Voss is a former FBI lead hostage negotiator and a leading authority on the art of negotiation. He teaches communication skills and strategies to help people get more of what they want every day. Some of his key principles of negotiation include showing the other side that you are negotiating in good faith, being genuinely interested in what drives the other side, taking emotions into consideration, building trust-based influence through the use of tactical empathy, working to deactivate negative feelings, aiming to magnify positive emotions, and keeping an eye out for black swans."""
    player4 = Player(name="Negotiation Expert", backend=OpenAIChat(model=MODEL),
                     role_desc=f"You are the negotiation expert like Chris Voss. Here is a brief description of Chris Voss:\n{christ_voss}",
                     global_prompt=environment_description)

    elon_musk = """Elon Musk is a visionary entrepreneur known for his views on technology and its potential to change the world. He has long been convinced that for life to survive, humanity has to become a multiplanet species. He founded Space Exploration Technologies (SpaceX) in 2002 to make more affordable rockets. Musk has also been involved in efforts to revolutionize battery technology. However, he has also warned of the dangers of artificial intelligence and has ramped up efforts in this area."""
    player5 = Player(name="Technology Expert", backend=OpenAIChat(model=MODEL),
                     role_desc=f"You are the technology expert like Elon Musk. Here is a brief description of Elon Musk:\n{elon_musk}",
                     global_prompt=environment_description)

    conversation = Conversation(
        player_names=[p.name for p in [ceo, player1, player2, player3, player4, player5]],
        # moderator=moderator,
        parallel=False,
        moderator_visibility="all",
        moderator_period="round",
    )

    arena = Arena(
        environment=conversation,
        players=[ceo, player1, player2, player3, player4, player5],
        global_prompt=environment_description,
    )
    arena.launch_cli(max_steps=100, interactive=True)


if __name__ == "__main__":
    main()



from typing import List, Union

from .base import TimeStep, Environment
from ..message import Message, MessagePool
from ..agent import Moderator, SIGNAL_END_OF_CONVERSATION
from ..config import EnvironmentConfig, AgentConfig


class Conversation(Environment):
    """
    Turn-based fully observable conversation environment.
    Next speaker order is either parallel or round-robin.
    """
    type_name = "conversation"

    def __init__(self, player_names: List[str], parallel: bool = False, **kwargs):
        super().__init__(player_names=player_names, parallel=parallel, **kwargs)

        self.parallel = parallel

        # The "state" of the environment is maintained by the message pool
        self.message_pool = MessagePool()

        self._current_turn = 0
        self._next_player_idx = 0

    def reset(self):
        self._current_turn = 0
        self._next_player_idx = 0
        self.message_pool.reset()

        init_timestep = TimeStep(observation=[],
                                 reward=self.get_zero_rewards(),
                                 terminal=False)
        return init_timestep

    def to_config(self) -> EnvironmentConfig:
        return EnvironmentConfig(env_type=self.type_name, player_names=self.player_names, parallel=self.parallel)

    def print(self):
        self.message_pool.print()

    def get_next_player(self) -> str:
        """
        get the next player
        """
        return self.player_names[self._next_player_idx]

    def get_observation(self, player_name=None) -> List[Message]:
        """
        get observation for the player
        """
        if player_name is None:
            return self.message_pool.get_all_messages()
        else:
            return self.message_pool.get_visible_messages(player_name, turn=self._current_turn)

    def is_terminal(self) -> bool:
        """
        check if the conversation is over
        """
        # If the last message is the signal, then the conversation is over
        if self.message_pool.last_message.content.startswith(SIGNAL_END_OF_CONVERSATION):
            return True

    def step(self, player_name: str, action: str) -> TimeStep:
        """
        step function that is called by the arena
        Args:
            player_name: the name of the player that takes the action
            action: the action that the agents wants to take
        """
        message = Message(agent_name=player_name, content=action, turn=self._current_turn)
        self.message_pool.append_message(message)

        # Update the counters
        if not self.parallel or self._next_player_idx == 0:
            self._current_turn += 1
        self._next_player_idx = (self._next_player_idx + 1) % self.num_players

        timestep = TimeStep(observation=self.get_observation(),
                            reward=self.get_zero_rewards(),
                            terminal=self.is_terminal())  # Return all the messages
        return timestep


class ModeratedConversation(Conversation):
    """
    Turn-based fully observable conversation environment.
    Next speaker order is either parallel or round-robin.
    Moderator is a special agent that can see all messages and can decide whether the conversation is over.
    """

    type_name = "moderated_conversation"

    def __init__(self, player_names: List[str], moderator: Union[Moderator, AgentConfig],
                 parallel: bool = False, moderator_visibility="all", moderator_period="turn", **kwargs):

        super().__init__(player_names=player_names, parallel=parallel, **kwargs)

        if isinstance(moderator, AgentConfig):
            moderator_config = moderator
            moderator = Moderator.from_config(moderator_config)
        elif not isinstance(moderator, Moderator):
            raise ValueError("moderator must be either an AgentConfig or a Moderator instance.")

        self.moderator = moderator
        self.moderator_visibility = moderator_visibility
        self.moderator_period = moderator_period

    def to_config(self) -> EnvironmentConfig:
        # This environment contains some speical config arguments that needs to be handle specially
        return EnvironmentConfig(env_type=self.type_name, player_names=self.player_names, parallel=self.parallel,
                                 moderator=self.moderator.to_config(), moderator_visibility=self.moderator_visibility,
                                 moderator_period=self.moderator_period)

    def step(self, player_name: str, action: str) -> TimeStep:
        """
        step function that is called by the arena
        Args:
            player_name: the name of the player that takes the action
            action: the action that the agents wants to take
        """
        message = Message(agent_name=player_name, content=action, turn=self._current_turn)
        self.message_pool.append_message(message)

        # Round-robin order for the next player
        self._next_player_idx = (self._next_player_idx + 1) % self.num_players

        if self.moderator_period == "turn" or \
                (self.moderator_period == "round" and self._next_player_idx == 0):
            # Moderator's turn
            moderator_history = self.message_pool.get_all_messages()
            moderator_response = self.moderator(moderator_history)
            moderator_message = Message(agent_name=self.moderator.name,
                                        content=moderator_response,
                                        turn=self._current_turn,
                                        visible_to=self.moderator_visibility)
            self.message_pool.append_message(moderator_message)
            terminal = self.moderator.is_terminal(moderator_history) or self.is_terminal()
        else:
            terminal = self.is_terminal()

        # Update the counters
        if not self.parallel or self._next_player_idx == 0:
            self._current_turn += 1

        timestep = TimeStep(observation=self.get_observation(),
                            reward=self.get_zero_rewards(),
                            terminal=terminal)  # Return all the messages
        return timestep
